/*
 * Copyright (C) the libgit2 contributors. All rights reserved.
 *
 * This file is part of libgit2, distributed under the GNU GPL v2 with
 * a Linking Exception. For full terms see the included COPYING file.
 */

#ifndef INCLUDE_stdalloc_h__
#define INCLUDE_stdalloc_h__

#include "alloc.h"

#include "common.h"

int git_stdalloc_init_allocator(git_allocator *allocator);

#endif
